/*
 *      dnsutl - utilities to make DNS easier to configure
 *      Copyright (C) 2006, 2007 Peter Miller
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see
 *      <http://www.gnu.org/licenses/>.
 */

#include <ac/stdio.h>

#include <conf.h>
#include <error.h>
#include <lex.h>
#include <mem.h>


conf_t *
conf_list_new(void)
{
    conf_t *cp = mem_alloc(sizeof(conf_t));
    cp->location = lex_location();
    cp->word = 0;
    cp->length = 0;
    cp->maximum = 0;
    cp->list = 0;
    return cp;
}


conf_t *
conf_string_new(string_ty *s)
{
    conf_t *cp = conf_list_new();
    cp->word = str_copy(s);
    return cp;
}


void
conf_delete(conf_t *cp)
{
    size_t          j;

    if (cp->word)
        str_free(cp->word);
    for (j = 0; j < cp->length; ++j)
        conf_delete(cp->list[j]);
    if (cp->list)
        mem_free(cp->list);
    mem_free(cp);
}


void
conf_list_append(conf_t *haystack, conf_t *needle)
{
    if (haystack->length >= haystack->maximum)
    {
        size_t          j;
        size_t          new_maximum;
        conf_t          **new_list;

        new_maximum = haystack->maximum * 2 + 8;
        new_list = mem_alloc(new_maximum * sizeof(haystack->list[0]));
        for (j = 0; j < haystack->length; ++j)
            new_list[j] = haystack->list[j];
        if (haystack->list)
            mem_free(haystack->list);
        haystack->list = new_list;
        haystack->maximum = new_maximum;
    }
    haystack->list[haystack->length++] = needle;
}


conf_t *
conf_find_a_name(conf_t *cp)
{
    size_t          j;

    if (cp->word)
        return cp;
    for (j = 0; j < cp->length; ++j)
    {
        conf_t          *result;

        result = conf_find_a_name(cp->list[j]);
        if (result)
            return result;
    }
    return 0;
}


void
conf_error(conf_t *cp, const char *fmt, ...)
{
    va_list ap;
    char buffer[1000];

    va_start(ap, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, ap);
    va_end(ap);
    error("%s: %s", cp->location->str_text, buffer);
}


string_ty *
conf_representation(conf_t *cp)
{
    string_ty *s;
    string_ty *s2;
    string_ty *s3;
    size_t j;

    if (cp->word)
        return str_format("\"%s\"", cp->word->str_text);
    s = str_from_c("{");
    for (j = 0; j < cp->length; ++j)
    {
        s2 = conf_representation(cp->list[j]);
        s3 = str_format("%s %s", s->str_text, s2->str_text);
        str_free(s);
        str_free(s2);
        s = s3;
    }
    s2 = str_format("%s }", s->str_text);
    str_free(s);
    return s2;
}
